"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.bundle_instance
hub.exec.boto3.client.ec2.describe_instances
hub.exec.boto3.client.ec2.import_instance
hub.exec.boto3.client.ec2.monitor_instances
hub.exec.boto3.client.ec2.reboot_instances
hub.exec.boto3.client.ec2.run_instances
hub.exec.boto3.client.ec2.start_instances
hub.exec.boto3.client.ec2.stop_instances
hub.exec.boto3.client.ec2.terminate_instances
hub.exec.boto3.client.ec2.unmonitor_instances
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", name)
hub.tool.boto3.resource.exec(resource, attach_classic_link_vpc, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_volume, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, console_output, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_image, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, describe_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_classic_link_vpc, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_volume, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, modify_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, monitor, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, password_data, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, reboot, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, report_status, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, reset_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, reset_kernel, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, reset_ramdisk, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, reset_source_dest_check, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, start, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, stop, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, terminate, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, unmonitor, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
            "aws.iam.role.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.subnet.present",
            "aws.iam.role.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    block_device_mappings: List = None,
    image_id: str = None,
    instance_type: str = None,
    ipv6_address_count: int = None,
    ipv6_addresses: List = None,
    kernel_id: str = None,
    key_name: str = None,
    monitoring: Dict = None,
    placement: Dict = None,
    ramdisk_id: str = None,
    security_group_ids: List = None,
    security_groups: List = None,
    subnet: str = None,
    user_data: str = None,
    additional_info: str = None,
    disable_api_termination: bool = None,
    ebs_optimized: bool = None,
    iam_instance_profile: Dict = None,
    instance_initiated_shutdown_behavior: str = None,
    network_interfaces: List = None,
    private_ip_address: str = None,
    elastic_gpu_specification: List = None,
    elastic_inference_accelerators: List = None,
    tags: Dict[str, str] = None,
    launch_template: Dict = None,
    instance_market_options: Dict = None,
    credit_specification: Dict = None,
    cpu_options: Dict = None,
    capacity_reservation_specification: Dict = None,
    hibernation_options: Dict = None,
    license_specifications: List = None,
    metadata_options: Dict = None,
    enclave_options: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Launches the specified number of instances using an AMI for which you have permissions. You can specify a number
    of options, or leave the default options. The following rules apply:   [EC2-VPC] If you don't specify a subnet
    ID, we choose a default subnet from your default VPC for you. If you don't have a default VPC, you must specify
    a subnet ID in the request.   [EC2-Classic] If don't specify an Availability Zone, we choose one for you.   Some
    instance types must be launched into a VPC. If you do not have a default VPC, or if you do not specify a subnet
    ID, the request fails. For more information, see Instance types available only in a VPC.   [EC2-VPC] All
    instances have a network interface with a primary private IPv4 address. If you don't specify this address, we
    choose one from the IPv4 range of your subnet.   Not all instance types support IPv6 addresses. For more
    information, see Instance types.   If you don't specify a security group ID, we use the default security group.
    For more information, see Security groups.   If any of the AMIs have a product code attached for which the user
    has not subscribed, the request fails.   You can create a launch template, which is a resource that contains the
    parameters to launch an instance. When you launch an instance using RunInstances, you can specify the launch
    template instead of specifying the launch parameters. To ensure faster instance launches, break up large
    requests into smaller batches. For example, create five separate launch requests for 100 instances each instead
    of one launch request for 500 instances. An instance is ready for you to use when it's in the running state. You
    can check the state of your instance using DescribeInstances. You can tag instances and EBS volumes during
    launch, after launch, or both. For more information, see CreateTags and Tagging your Amazon EC2 resources. Linux
    instances have access to the public key of the key pair at boot. You can use this key to provide secure access
    to the instance. Amazon EC2 public images use this feature to provide secure access without passwords. For more
    information, see Key pairs. For troubleshooting, see What to do if an instance immediately terminates, and
    Troubleshooting connecting to your instance.

    Args:
        name(Text): A name, ID, or JMES search path to identify the resource.
        block_device_mappings(List, optional): The block device mapping entries. Defaults to None.
        image_id(Text, optional): The ID of the AMI. An AMI ID is required to launch an instance and must be specified here or in
            a launch template. Defaults to None.
        instance_type(Text, optional): The instance type. For more information, see Instance types in the Amazon EC2 User Guide.
            Default: m1.small. Defaults to None.
        ipv6_address_count(int, optional): [EC2-VPC] The number of IPv6 addresses to associate with the primary network interface. Amazon
            EC2 chooses the IPv6 addresses from the range of your subnet. You cannot specify this option and
            the option to assign specific IPv6 addresses in the same request. You can specify this option if
            you've specified a minimum number of instances to launch. You cannot specify this option and the
            network interfaces option in the same request. Defaults to None.
        ipv6_addresses(List, optional): [EC2-VPC] The IPv6 addresses from the range of the subnet to associate with the primary network
            interface. You cannot specify this option and the option to assign a number of IPv6 addresses in
            the same request. You cannot specify this option if you've specified a minimum number of
            instances to launch. You cannot specify this option and the network interfaces option in the
            same request. Defaults to None.
        kernel_id(Text, optional): The ID of the kernel.  We recommend that you use PV-GRUB instead of kernels and RAM disks. For
            more information, see  PV-GRUB in the Amazon EC2 User Guide. Defaults to None.
        key_name(Text, optional): The name of the key pair. You can create a key pair using CreateKeyPair or ImportKeyPair.  If
            you do not specify a key pair, you can't connect to the instance unless you choose an AMI that
            is configured to allow users another way to log in. Defaults to None.
        monitoring(Dict, optional): Specifies whether detailed monitoring is enabled for the instance. Defaults to None.
        placement(Dict, optional): The placement for the instance. Defaults to None.
        ramdisk_id(Text, optional): The ID of the RAM disk to select. Some kernels require additional drivers at launch. Check the
            kernel requirements for information about whether you need to specify a RAM disk. To find kernel
            requirements, go to the Amazon Web Services Resource Center and search for the kernel ID.  We
            recommend that you use PV-GRUB instead of kernels and RAM disks. For more information, see  PV-
            GRUB in the Amazon EC2 User Guide. Defaults to None.
        security_group_ids(List, optional): The IDs of the security groups. You can create a security group using CreateSecurityGroup. If
            you specify a network interface, you must specify any security groups as part of the network
            interface. Defaults to None.
        security_groups(List, optional): [EC2-Classic, default VPC] The names of the security groups. For a nondefault VPC, you must use
            security group IDs instead. If you specify a network interface, you must specify any security
            groups as part of the network interface. Default: Amazon EC2 uses the default security group. Defaults to None.
        subnet(Text, optional): [EC2-VPC] The ID of the subnet to launch the instance into. If you specify a network interface,
            you must specify any subnets as part of the network interface. Defaults to None.
        user_data(Text, optional): The user data to make available to the instance. For more information, see Running commands on
            your Linux instance at launch (Linux) and Adding User Data (Windows). If you are using a command
            line tool, base64-encoding is performed for you, and you can load the text from a file.
            Otherwise, you must provide base64-encoded text. User data is limited to 16 KB. Defaults to None.
        additional_info(Text, optional): Reserved. Defaults to None.
        disable_api_termination(bool, optional): If you set this parameter to true, you can't terminate the instance using the Amazon EC2
            console, CLI, or API; otherwise, you can. To change this attribute after launch, use
            ModifyInstanceAttribute. Alternatively, if you set InstanceInitiatedShutdownBehavior to
            terminate, you can terminate the instance by running the shutdown command from the instance.
            Default: false. Defaults to None.
        ebs_optimized(bool, optional): Indicates whether the instance is optimized for Amazon EBS I/O. This optimization provides
            dedicated throughput to Amazon EBS and an optimized configuration stack to provide optimal
            Amazon EBS I/O performance. This optimization isn't available with all instance types.
            Additional usage charges apply when using an EBS-optimized instance. Default: false. Defaults to None.
        iam_instance_profile(Dict, optional): The name or Amazon Resource Name (ARN) of an IAM instance profile. Defaults to None.
        instance_initiated_shutdown_behavior(Text, optional): Indicates whether an instance stops or terminates when you initiate shutdown from the instance
            (using the operating system command for system shutdown). Default: stop. Defaults to None.
        network_interfaces(List, optional): The network interfaces to associate with the instance. If you specify a network interface, you
            must specify any security groups and subnets as part of the network interface. Defaults to None.
        private_ip_address(Text, optional): [EC2-VPC] The primary IPv4 address. You must specify a value from the IPv4 address range of the
            subnet. Only one private IP address can be designated as primary. You can't specify this option
            if you've specified the option to designate a private IP address as the primary IP address in a
            network interface specification. You cannot specify this option if you're launching more than
            one instance in the request. You cannot specify this option and the network interfaces option in
            the same request. Defaults to None.
        elastic_gpu_specification(List, optional): An elastic GPU to associate with the instance. An Elastic GPU is a GPU resource that you can
            attach to your Windows instance to accelerate the graphics performance of your applications. For
            more information, see Amazon EC2 Elastic GPUs in the Amazon EC2 User Guide. Defaults to None.
        elastic_inference_accelerators(List, optional): An elastic inference accelerator to associate with the instance. Elastic inference accelerators
            are a resource you can attach to your Amazon EC2 instances to accelerate your Deep Learning (DL)
            inference workloads. You cannot specify accelerators from different generations in the same
            request. Defaults to None.
        tags(Dict, optional): The tags to apply to the resources during launch. You can only tag instances and volumes on
            launch. The specified tags are applied to all instances or volumes that are created during
            launch. To tag a resource after it has been created, see CreateTags. Defaults to None.
        launch_template(Dict, optional): The launch template to use to launch the instances. Any parameters that you specify in
            RunInstances override the same parameters in the launch template. You can specify either the
            name or ID of a launch template, but not both. Defaults to None.
        instance_market_options(Dict, optional): The market (purchasing) option for the instances. For RunInstances, persistent Spot Instance
            requests are only supported when InstanceInterruptionBehavior is set to either hibernate or
            stop. Defaults to None.
        credit_specification(Dict, optional): The credit option for CPU usage of the burstable performance instance. Valid values are standard
            and unlimited. To change this attribute after launch, use  ModifyInstanceCreditSpecification.
            For more information, see Burstable performance instances in the Amazon EC2 User Guide. Default:
            standard (T2 instances) or unlimited (T3/T3a instances). Defaults to None.
        cpu_options(Dict, optional): The CPU options for the instance. For more information, see Optimizing CPU options in the Amazon
            EC2 User Guide. Defaults to None.
        capacity_reservation_specification(Dict, optional): Information about the Capacity Reservation targeting option. If you do not specify this
            parameter, the instance's Capacity Reservation preference defaults to open, which enables it to
            run in any open Capacity Reservation that has matching attributes (instance type, platform,
            Availability Zone). Defaults to None.
        hibernation_options(Dict, optional): Indicates whether an instance is enabled for hibernation. For more information, see Hibernate
            your instance in the Amazon EC2 User Guide. You can't enable hibernation and Amazon Web Services
            Nitro Enclaves on the same instance. Defaults to None.
        license_specifications(List, optional): The license configurations. Defaults to None.
        metadata_options(Dict, optional): The metadata options for the instance. For more information, see Instance metadata and user
            data. Defaults to None.
        enclave_options(Dict, optional): Indicates whether the instance is enabled for Amazon Web Services Nitro Enclaves. For more
            information, see  What is Amazon Web Services Nitro Enclaves? in the Amazon Web Services Nitro
            Enclaves User Guide. You can't enable Amazon Web Services Nitro Enclaves and hibernation on the
            same instance. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws_auto.ec2.instance.present:
                - name: value
    """
    result = dict(comment="", changes=None, name=name, result=True)
    ret = await hub.exec.boto3.client.ec2.instance.search(ctx, jmes_path=name)

    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    ret = await hub.exec.boto3.client.ec2.subnet.search(ctx, jmes_path=subnet)
    if ret["result"]:
        # name is now the first id that matched the JMES search path
        subnet = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if before:
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.run_instances(
                ctx,
                DryRun=ctx.test,
                ClientToken=name,
                **{
                    "BlockDeviceMappings": block_device_mappings,
                    "ImageId": image_id,
                    "InstanceType": instance_type,
                    "Ipv6AddressCount": ipv6_address_count,
                    "Ipv6Addresses": ipv6_addresses,
                    "KernelId": kernel_id,
                    "KeyName": key_name,
                    "MaxCount": 1,
                    "MinCount": 1,
                    "Monitoring": monitoring,
                    "Placement": placement,
                    "RamdiskId": ramdisk_id,
                    "SecurityGroupIds": security_group_ids,
                    "SecurityGroups": security_groups,
                    "SubnetId": subnet,
                    "UserData": user_data,
                    "AdditionalInfo": additional_info,
                    "DisableApiTermination": disable_api_termination,
                    "EbsOptimized": ebs_optimized,
                    "IamInstanceProfile": iam_instance_profile,
                    "InstanceInitiatedShutdownBehavior": instance_initiated_shutdown_behavior,
                    "NetworkInterfaces": network_interfaces,
                    "PrivateIpAddress": private_ip_address,
                    "ElasticGpuSpecification": elastic_gpu_specification,
                    "ElasticInferenceAccelerators": elastic_inference_accelerators,
                    "LaunchTemplate": launch_template,
                    "InstanceMarketOptions": instance_market_options,
                    "CreditSpecification": credit_specification,
                    "CpuOptions": cpu_options,
                    "CapacityReservationSpecification": capacity_reservation_specification,
                    "HibernationOptions": hibernation_options,
                    "LicenseSpecifications": license_specifications,
                    "MetadataOptions": metadata_options,
                    "EnclaveOptions": enclave_options,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            name = ret["ret"]["Instances"][0]["InstanceId"]
            resource._id = name
            result["comment"] = f"Created '{name}'"
            if tags:
                resource.create_tags(
                    DryRun=ctx.test,
                    Tags=[{"Key": k, "Value": v} for k, v in tags.items()],
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    # TODO perform other modifications as needed here
    ...

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["changes"] = differ.deep_diff(before, after)
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Shuts down the specified instances. This operation is idempotent; if you terminate an instance more than once,
    each call succeeds.  If you specify multiple instances and the request fails (for example, because of a single
    incorrect instance ID), none of the instances are terminated. If you terminate multiple instances across
    multiple Availability Zones, and one or more of the specified instances are enabled for termination protection,
    the request fails with the following results:   The specified instances that are in the same Availability Zone
    as the protected instance are not terminated.   The specified instances that are in different Availability
    Zones, where no other specified instances are protected, are successfully terminated.   For example, say you
    have the following instances:   Instance A: us-east-1a; Not protected   Instance B: us-east-1a; Not protected
    Instance C: us-east-1b; Protected   Instance D: us-east-1b; not protected   If you attempt to terminate all of
    these instances in the same request, the request reports failure with the following results:   Instance A and
    Instance B are successfully terminated because none of the specified instances in us-east-1a are enabled for
    termination protection.   Instance C and Instance D fail to terminate because at least one of the specified
    instances in us-east-1b (Instance C) is enabled for termination protection.   Terminated instances remain
    visible after termination (for approximately one hour). By default, Amazon EC2 deletes all EBS volumes that were
    attached when the instance launched. Volumes attached after instance launch continue running. You can stop,
    start, and terminate EBS-backed instances. You can only terminate instance store-backed instances. What happens
    to an instance differs if you stop it or terminate it. For example, when you stop an instance, the root device
    and any other devices attached to the instance persist. When you terminate an instance, any attached EBS volumes
    with the DeleteOnTermination block device mapping parameter set to true are automatically deleted. For more
    information about the differences between stopping and terminating instances, see Instance lifecycle in the
    Amazon EC2 User Guide. For more information about troubleshooting, see Troubleshooting terminating your instance
    in the Amazon EC2 User Guide.

    Args:
        name(Text): The id of the instance

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.instance.absent:
                - name: value
                - instance_ids: value
    """

    result = dict(comment="", changes=None, name=name, result=True)

    ret = await hub.exec.boto3.client.ec2.instance.search(ctx, jmes_path=name)

    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Instance", name)

    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' already absent"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.terminate_instances(
                ctx, DryRun=ctx.test, **{"InstanceIds": [name]}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    after = await hub.tool.boto3.resource.describe(resource)
    result["changes"] = differ.deep_diff(before, after)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_instances(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe Instances {ret['comment']}")
        return {}

    instances = []
    for reservation in ret["ret"]["Reservations"]:
        instances.extend(reservation["Instances"])
    for instance in instances:
        instance_id = instance["InstanceId"]
        new_instance = [
            {"block_device_mappings": instance["BlockDeviceMappings"]},
            {"image_id": instance["ImageId"]},
            {"instance_type": instance["InstanceType"]},
            {"kernel_id": instance["KernelId"]},
            {"key_name": instance["KeyName"]},
            {"monitoring": instance["Monitoring"]},
            {"placement": instance["Placement"]},
            {"subnet": instance["SubnetId"]},
            {"ebs_optimized": instance["EbsOptimized"]},
            {"private_ip_address": instance["PrivateIpAddress"]},
            {"tags": {item["Key"]: item["Value"] for item in instance.get("Tags", [])}},
            # TODO Get all this too
            # ipv6_address_count=None,
            # ramdisk_id=None,
            # security_group_ids=None,
            # security_groups=None,
            # user_data=None,
            # additional_info=None,
            # disable_api_termination=None,
            # iam_instance_profile=None,
            # instance_initiated_shutdown_behavior=None,
            # network_interfaces=None,
            # elastic_gpu_specification=None,
            # elastic_inference_accelerators=None,
            # launch_template=None,
            # instance_market_options=None,
            # credit_specification=None,
            # cpu_options=None,
            # capacity_reservation_specification=None,
            # hibernation_options=None,
            # license_specifications=None,
            # metadata_options=None,
            # enclave_options=None,
        ]

        result[instance_id] = {"aws.ec2.instance.present": new_instance}
        for i, data in enumerate(instance["NetworkInterfaces"]):
            sub_instance = copy.deepcopy(new_instance)
            sub_instance.append({"ipv6_addresses": data.get("Ipv6Addresses", [])})
            # The id needs to be in the name
            sub_instance.append({"name": instance["InstanceId"]})
            # Create a new state for this association set
            result[f"{instance_id}-{i}"] = {"aws.ec2.instance.present": sub_instance}
    return result



import botocore.client
import botocore.docs.docstring
import botocore.exceptions
import botocore.model

__func_alias__ = {"type_": "type"}


def parse(hub, param: "botocore.model.Shape", required: bool):
    docstring = hub.tool.format.html.parse(param.documentation)
    return {
        "required": required,
        "default": None,
        "target_type": "mapping",
        "target": "kwargs",
        "param_type": hub.pop_create.aws.plugin.type(param.type_name),
        "doc": "\n            ".join(
            hub.tool.format.wrap.wrap(
                docstring,
                width=96,
            )
        ),
    }


def type_(hub, type_name: str):
    if type_name == "string":
        return "Text"
    elif type_name == "map":
        return "Dict"
    elif type_name == "structure":
        return "Dict"
    elif type_name == "list":
        return "List"
    elif type_name == "boolean":
        return "bool"
    elif type_name in ("integer", "long"):
        return "int"
    elif type_name in ("float", "double"):
        return "float"
    elif type_name == "timestamp":
        return "Text"
    elif type_name == "blob":
        return "ByteString"
    else:
        raise NameError(type_name)



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_vpc
hub.exec.boto3.client.ec2.delete_vpc
hub.exec.boto3.client.ec2.describe_vpcs
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
hub.tool.boto3.resource.exec(resource, associate_dhcp_options, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, attach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_network_acl, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route_table, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_security_group, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, describe_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_classic_link_instance, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, detach_internet_gateway, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, disable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, enable_classic_link, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, modify_attribute, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, request_vpc_peering_connection, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    cidr_block: str,
    amazon_provided_ipv6_cidr_block: bool = None,
    ipv6_pool: str = None,
    ipv6_cidr_block: str = None,
    instance_tenancy: str = None,
    ipv6_cidr_block_network_border_group: str = None,
    tag_specifications: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a VPC with the specified IPv4 CIDR block. The smallest VPC you can create uses a /28 netmask (16 IPv4
    addresses), and the largest uses a /16 netmask (65,536 IPv4 addresses). For more information about how large to
    make your VPC, see Your VPC and subnets in the Amazon Virtual Private Cloud User Guide. You can optionally
    request an IPv6 CIDR block for the VPC. You can request an Amazon-provided IPv6 CIDR block from Amazon's pool of
    IPv6 addresses, or an IPv6 CIDR block from an IPv6 address pool that you provisioned through bring your own IP
    addresses (BYOIP). By default, each instance you launch in the VPC has the default DHCP options, which include
    only a default DNS server that we provide (AmazonProvidedDNS). For more information, see DHCP options sets in
    the Amazon Virtual Private Cloud User Guide. You can specify the instance tenancy value for the VPC when you
    create it. You can't change this value for the VPC after you create it. For more information, see Dedicated
    Instances in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name(Text): A name, ID, or JMES search path to identify the resource.
        cidr_block(Text): The IPv4 network range for the VPC, in CIDR notation. For example, 10.0.0.0/16. We modify the
            specified CIDR block to its canonical form; for example, if you specify 100.68.0.18/18, we
            modify it to 100.68.0.0/18.
        amazon_provided_ipv6_cidr_block(bool, optional): Requests an Amazon-provided IPv6 CIDR block with a /56 prefix length for the VPC. You cannot
            specify the range of IP addresses, or the size of the CIDR block. Defaults to None.
        ipv6_pool(Text, optional): The ID of an IPv6 address pool from which to allocate the IPv6 CIDR block. Defaults to None.
        ipv6_cidr_block(Text, optional): The IPv6 CIDR block from the IPv6 address pool. You must also specify Ipv6Pool in the request.
            To let Amazon choose the IPv6 CIDR block for you, omit this parameter. Defaults to None.
        instance_tenancy(Text, optional): The tenancy options for instances launched into the VPC. For default, instances are launched
            with shared tenancy by default. You can launch instances with any tenancy into a shared tenancy
            VPC. For dedicated, instances are launched as dedicated tenancy instances by default. You can
            only launch instances with a tenancy of dedicated or host into a dedicated tenancy VPC.
            Important: The host value cannot be used with this parameter. Use the default or dedicated
            values only. Default: default. Defaults to None.
        ipv6_cidr_block_network_border_group(Text, optional): The name of the location from which we advertise the IPV6 CIDR block. Use this parameter to
            limit the address to this location.  You must set AmazonProvidedIpv6CidrBlock to true to use
            this parameter. Defaults to None.
        tag_specifications(List, optional): The tags to assign to the VPC. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws_auto.ec2.vpc.present:
                - name: value
                - cidr_block: value
    """
    result = dict(comment="", changes=None, name=name, result=True)
    ret = await hub.exec.boto3.client.ec2.vpc.search(ctx, jmes_path=name)
    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if before:
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_vpc(
                ctx,
                DryRun=ctx.test,
                **{
                    "CidrBlock": cidr_block,
                    "AmazonProvidedIpv6CidrBlock": amazon_provided_ipv6_cidr_block,
                    "Ipv6Pool": ipv6_pool,
                    "Ipv6CidrBlock": ipv6_cidr_block,
                    "InstanceTenancy": instance_tenancy,
                    "Ipv6CidrBlockNetworkBorderGroup": ipv6_cidr_block_network_border_group,
                    "TagSpecifications": tag_specifications,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            ret["comment"] = f"Created '{name}'"

            if ret["result"]:
                resource._id = ret["ret"]["Vpc"]["VpcId"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    # TODO perform other modifications as needed here
    ...

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["changes"] = differ.deep_diff(before, after)
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC. You must detach or delete all gateways and resources that are associated with the VPC
    before you can delete it. For example, you must terminate all instances running in the VPC, delete all security
    groups associated with the VPC (except the default one), delete all route tables associated with the VPC (except
    the default one), and so on.

    Args:
        name(Text): The ID of the VPC or a name, ID, or JMES search path to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.vpc.absent:
                - name: value
                - vpc_id: value
    """

    result = dict(comment="", changes=None, name=name, result=True)
    ret = await hub.exec.boto3.client.ec2.vpc.search(ctx, jmes_path=name)
    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Vpc", name)

    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' already absent"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.delete_vpc(
                ctx, DryRun=ctx.test, **{"VpcId": name}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["changes"] = differ.deep_diff(before, after)
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpcs(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe VPCs {ret['comment']}")
        return {}

    for vpc in ret["ret"]["Vpcs"]:
        new_vpc = [
            {"cidr_block": vpc["CidrBlock"]},
            {
                "amazon_provided_ipv6_cidr_block": bool(
                    vpc["Ipv6CidrBlockAssociationSet"]
                )
            },
            {"instance_tenancy": vpc["InstanceTenancy"]},
            {"tag_specifications": vpc["Tags"]},
        ]
        result[vpc["VpcId"]] = {"aws.ec2.vpc.present": new_vpc}
        for i, data in enumerate(vpc["Ipv6CidrBlockAssociationSet"]):
            sub_vpc = copy.deepcopy(new_vpc)
            sub_vpc.append({"ipv6_cidr_block": data["Ipv6CidrBlock"]})
            sub_vpc.append(
                {"ipv6_cidr_block_network_border_group": data["NetworkBorderGroup"]}
            )
            sub_vpc.append({"ipv6_pool": data["Ipv6Pool"]})
            # The id needs to be in the name
            sub_vpc.append({"name": vpc["VpcId"]})
            # Create a new state for this association set
            result[f"{vpc['VpcId']}-{i}"] = {"aws.ec2.vpc.present": sub_vpc}
    return result



import botocore.client
import botocore.docs.docstring


def parse(hub, client: "botocore.client.BaseClient", service: str, operation: str):
    function = getattr(client, operation)
    doc: botocore.docs.docstring.ClientMethodDocstring = function.__doc__
    docstring = hub.tool.format.html.parse(doc._gen_kwargs["method_description"])
    try:
        # TODO what does botocore expect?
        params = doc._gen_kwargs["operation_model"].input_shape.members
        required_params = doc._gen_kwargs[
            "operation_model"
        ].input_shape.required_members
        parameters = {
            p: hub.pop_create.aws.plugin.parameter(
                param=data, required=p in required_params
            )
            for p, data in params.items()
        }
    except AttributeError:
        parameters = {}
    try:
        return_type = hub.pop_create.aws.plugin.type(
            doc._gen_kwargs["operation_model"].output_shape.type_name
        )
    except AttributeError:
        return_type = None

    ret = {
        "doc": "\n".join(hub.tool.format.wrap.wrap(docstring, width=112)),
        "params": parameters,
        "return_type": return_type,
        "hardcoded": {"service": service, "operation": operation},
    }

    return ret



from typing import Any
from typing import Dict

import dict_tools.data

__virtualname__ = "saml"


async def gather(hub) -> Dict[str, Any]:
    """
    Get profile names from encrypted AWS credential files.
    Standardize the profile keys to what a boto3 Client requires.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sts.html#STS.Client.assume_role_with_saml

    Example:
    .. code-block:: yaml

        aws.saml:
          profile_name:
            region: us-east-1
            role_arn: string
            principal_arn: string
            policy_arns:
              - arn: string1
              - arn: string2
            policy: string
            duration: 180
    """
    sub_profiles = {}
    for profile, ctx in hub.acct.PROFILES.get("aws.saml", {}).items():
        hub.tool.boto3.session.get()
        temp_ctx = dict_tools.data.NamespaceDict(
            acct={"region_name": ctx.get("region")}
        )
        saml_xml = "TODO"
        ret = await hub.tool.boto3.client.exec(
            temp_ctx,
            "sts",
            "assume_role_with_saml",
            RoleArn="arn:aws:sts::883373499178:assumed-role/developer/tyjohnson@vmware.com",
            PrincipalArn="arn:aws:sts::883373499178:assumed-role/developer/tyjohnson@vmware.com",
            SAMLAssertion=saml_xml,
        )
        sub_profiles[profile] = {
            "aws_access_key_id": ret["Credentials"]["AccessKeyId"],
            "aws_secret_access_key": ret["Credentials"]["SecretAccessKey"],
            "aws_session_token": ret["Credentials"]["SessionToken"],
            **ctx,
        }
    print(sub_profiles)
    return sub_profiles



import boto3.session
import botocore.session


USE_SHARED_BOTOCORE_SESSION = object()


def __init__(hub):
    # Create a single session for everything else to be run from
    hub.tool.boto3.SESSION = botocore.session.Session()


def get(
    hub, botocore_session: botocore.session.Session = USE_SHARED_BOTOCORE_SESSION
) -> boto3.session.Session:
    """
    Get a boto3 session that uses the universal botocore session
    """
    if botocore_session == USE_SHARED_BOTOCORE_SESSION:
        return boto3.session.Session(botocore_session=hub.tool.boto3.SESSION)
    else:
        return boto3.session.Session(botocore_session=botocore_session)



DESCRIBE_FUNCTIONS = ("describe", "get", "search", "list")

DELETE_FUNCTIONS = (
    "delete",
    "disassociate",
    "reject",
    "deallocate",
    "unassign",
    "deregister",
    "deprovision",
    "revoke",
    "release",
    "terminate",
    "cancel",
    "disable",
)

CREATE_FUNCTIONS = (
    "create",
    "associate",
    "accept",
    "allocate",
    "assign",
    "register",
    "provision",
    "authorize",
    "run",
    "enable",
    "upload",
    "put",
    "publish",
)

NAME_PARAMETER = {
    "default": None,
    "doc": "A name, ID, or JMES search path to identify the resource",
    "param_type": "Text",
    "required": True,
    "target": "hardcoded",
    "target_type": "arg",
}


PRESENT_REQUEST_FORMAT = r"""
    result = dict(comment="", changes= None, name=name, result=True)

    {%- if (not function.hardcoded.is_idempotent) %}
    ret = await hub.exec.boto3.client.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }}.id(
        ctx,
        jmes_path=name
    )
    if ret["status"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]
    {% endif %}

    {{ function.hardcoded.resource_function_call }}
    before = {{ function.hardcoded.describe_function_call }}

    if before:
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await {{ function.hardcoded.create_function }}(
                ctx,
                {{ "DryRun=ctx.test," if function.hardcoded.has_dry_run }}
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["status"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            ret["comment"] = f"Created '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    {{ function.hardcoded.waiter_call }}
    # TODO perform other modifications as needed here
    ...

    after = {{ function.hardcoded.describe_function_call }}
    result["changes"] = differ.deep_diff(before, after)
    return result
"""

ABSENT_REQUEST_FORMAT = r"""
    result = dict(comment="", changes= None, name=name, result=True)
    {%- if (not function.hardcoded.is_idempotent) %}
    ret = await hub.exec.boto3.client.{{ function.hardcoded.service_name }}.{{ function.hardcoded.resource }}.id(
        ctx,
        jmes_path=name
    )
    if ret["status"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]
    {% endif %}

    {{ function.hardcoded.resource_function_call }}

    before = {{ function.hardcoded.describe_function_call }}

    if not before:
        result["comment"] = f"'{name}' already absent"
    else:
        try:
            ret = await {{ function.hardcoded.delete_function }}(
                ctx,
                {{ "DryRun=ctx.test," if function.hardcoded.has_dry_run }}
                {{ "ClientToken=name," if function.hardcoded.has_client_token }}
                **{{ parameter.mapping.kwargs|default({}) }}
            )
            result["result"] = ret["status"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    {{ function.hardcoded.waiter_call }}

    after = {{ function.hardcoded.describe_function_call }}
    result["changes"] = differ.deep_diff(before, after)
    return result
"""

DESCRIBE_REQUEST_FORMAT = r"""
    result = {}

    ret = await {{ function.hardcoded.list_function}}(ctx)
    if not ret["status"]:
        hub.log.debug(f"Could not describe {{ function.hardcoded.resource }} {ret['comment']}")
        return result

    for {{ function.hardcoded.resource }} in ret["ret"]["{{ function.hardcoded.list_item }}"]:
        new_{{ function.hardcoded.resource }} = [
            {{ function.hardcoded.present_params }}
        ]
        result[{{ function.hardcoded.resource }}["{{ function.hardcoded.resource }}Id"]] = {"{{ function.ref }}.present": new_{{ function.hardcoded.resource }}}

        for i, data in enumerate({{ function.hardcoded.resource }}.get("", ())):
            sub_{{ function.hardcoded.resource }} = copy.deepcopy(new_{{ function.hardcoded.resource }})

            # TODO check for subresouruces
            sub_{{ function.hardcoded.resource }}.append({})
            sub_{{ function.hardcoded.resource }}.append({"name": "{{ function.hardcoded.resource_id }}"})
            result[f"{{ function.hardcoded.resource_id }}-{i}"] = {"{{ function.ref }}.present": sub_{{ function.hardcoded.resource }}}

    return result
"""



# curl 169.254.169.254/latest/meta-data/identity-credentials/ec2/security-credentials/
# Do it with V1 and V2, have "latest" be the default but let that value be configurable


def gather(hub):
    ...



from typing import Any
from typing import Dict

from dict_tools import differ

SERVICE = "iam"
RESOURCE = "User"


async def absent(hub, ctx, name: str, **kwargs) -> Dict[str, Any]:
    resource = hub.tool.boto3.resource.create(ctx, SERVICE, RESOURCE, name=name)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        comment = f"{RESOURCE} '{name}' already absent"
    else:
        try:
            # Attempt to delete the User
            await hub.tool.boto3.resource.exec(resource, "delete", **kwargs)
            comment = f"Delete {RESOURCE} '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            # User already deleted
            comment = f"{e.__class__.__name__: {e}}"

    after = await hub.tool.boto3.resource.describe(resource)
    return {
        "comment": comment,
        "changes": differ.deep_diff(before, after),
        "name": name,
        "result": not bool(after),
    }


async def present(hub, ctx, name: str, **kwargs) -> Dict[str, Any]:
    resource = hub.tool.boto3.resource.create(ctx, SERVICE, RESOURCE, name=name)
    before = await hub.tool.boto3.resource.describe(resource)

    if before:
        comment = f"{RESOURCE} '{name}' already exists"
    else:
        try:
            # Attempt to create the User
            await hub.tool.boto3.resource.exec(resource, "create", **kwargs)
            comment = f"Created {RESOURCE}, '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            # User already exists
            comment = f"{e.__class__.__name__: {e}}"

        await hub.tool.boto3.client.wait(ctx, SERVICE, "user_exists", UserName=name)

    # TODO perform other updates as necessary

    after = await hub.tool.boto3.resource.describe(resource)
    return {
        "comment": comment,
        "changes": differ.deep_diff(before, after),
        "name": name,
        "result": bool(after),
    }



import configparser


"""
https://blog.gruntwork.io/authenticating-to-aws-with-the-credentials-file-d16c0fbcbf9e

$ aws configure
AWS Access Key ID: AKIAIOSFODNN7EXAMPLE
AWS Secret Access Key: wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY
Default region name [None]: us-west-2
Default output format [None]: json

AWS prompts you to enter your Access Key ID and Secret Access Key and stores them in ~/.aws/credentials:

[default]
aws_access_key_id=AKIAIOSFODNN7EXAMPLE
aws_secret_access_key=wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY

It also stores the other settings you entered in ~/.aws/config:

[default]
region=us-west-2
output=json
"""


async def gather(hub, profiles):
    """
    load profiles from unencrypted AWS credential files

    Example:
    .. code-block:: yaml

        aws.iam:
            paths:
                - ~/.aws/credentials
                - /path/to/other/credential/file
            # Optional overrides
            id: XXXXXXXXXXXXXXXXX
            key: XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            region: us-east-1
    """
    sub_profiles = {}
    config = configparser.ConfigParser()

    ctx = profiles.get("aws.iam", {})
    credential_files = ctx.pop("paths", [])
    config.read(credential_files)

    for profile in set(config.sections()):
        sub_profiles[profile] = ctx

    return sub_profiles



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_subnet
hub.exec.boto3.client.ec2.delete_subnet
hub.exec.boto3.client.ec2.describe_subnets
resource = hub.tool.boto3.resource.create(ctx, "ec2", "Subnet", name)
hub.tool.boto3.resource.exec(resource, create_instances, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_network_interface, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.ec2.vpc.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.vpc.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    vpc: str,
    cidr_block: str,
    tag_specifications: List = None,
    availability_zone: str = None,
    availability_zone_id: str = None,
    ipv6_cidr_block: str = None,
    outpost_arn: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a subnet in a specified VPC. You must specify an IPv4 CIDR block for the subnet. After you create a
    subnet, you can't change its CIDR block. The allowed block size is between a /16 netmask (65,536 IP addresses)
    and /28 netmask (16 IP addresses). The CIDR block must not overlap with the CIDR block of an existing subnet in
    the VPC. If you've associated an IPv6 CIDR block with your VPC, you can create a subnet with an IPv6 CIDR block
    that uses a /64 prefix length.   Amazon Web Services reserves both the first four and the last IPv4 address in
    each subnet's CIDR block. They're not available for use.  If you add more than one subnet to a VPC, they're set
    up in a star topology with a logical router in the middle. When you stop an instance in a subnet, it retains its
    private IPv4 address. It's therefore possible to have a subnet with no running instances (they're all stopped),
    but no remaining IP addresses available. For more information about subnets, see Your VPC and subnets in the
    Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): A name, ID, or JMES search path to identify the resource.
        tag_specifications(List, optional): The tags to assign to the subnet. Defaults to None.
        availability_zone(Text, optional): The Availability Zone or Local Zone for the subnet. Default: Amazon Web Services selects one for
            you. If you create more than one subnet in your VPC, we do not necessarily select a different
            zone for each subnet. To create a subnet in a Local Zone, set this value to the Local Zone ID,
            for example us-west-2-lax-1a. For information about the Regions that support Local Zones, see
            Available Regions in the Amazon Elastic Compute Cloud User Guide. To create a subnet in an
            Outpost, set this value to the Availability Zone for the Outpost and specify the Outpost ARN. Defaults to None.
        availability_zone_id(Text, optional): The AZ ID or the Local Zone ID of the subnet. Defaults to None.
        ipv6_cidr_block(Text, optional): The IPv6 network range for the subnet, in CIDR notation. The subnet size must use a /64 prefix
            length. Defaults to None.
        outpost_arn(Text, optional): The Amazon Resource Name (ARN) of the Outpost. If you specify an Outpost ARN, you must also
            specify the Availability Zone of the Outpost subnet. Defaults to None.
        vpc(Text): JMES search path or ID of the VPC.
        cidr_block(Text): The IPv4 network range for the subnet, in CIDR notation. For example, 10.0.0.0/24. We modify the
            specified CIDR block to its canonical form; for example, if you specify 100.68.0.18/18, we
            modify it to 100.68.0.0/18.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws_auto.ec2.subnet.present:
                - name: value
                - vpc: value
                - cidr_block: value
    """

    result = dict(comment="", changes=None, name=name, result=True)
    ret = await hub.exec.boto3.client.ec2.subnet.search(ctx, jmes_path=name)
    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    ret = await hub.exec.boto3.client.ec2.vpc.search(ctx, jmes_path=vpc)
    if ret["result"]:
        vpc = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Subnet", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if before:
        result["comment"] = f"'{name}' already exists"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.create_subnet(
                ctx,
                DryRun=ctx.test,
                **{
                    "TagSpecifications": tag_specifications,
                    "AvailabilityZone": availability_zone,
                    "AvailabilityZoneId": availability_zone_id,
                    "Ipv6CidrBlock": ipv6_cidr_block,
                    "OutpostArn": outpost_arn,
                    "VpcId": vpc,
                    "CidrBlock": cidr_block,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            ret["comment"] = f"Created '{name}'"
            ret = await hub.exec.boto3.client.ec2.subnet.search(ctx, jmes_path=name)
            if ret["result"]:
                resource._id = ret["ret"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    # TODO perform other modifications as needed here
    ...

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["changes"] = differ.deep_diff(before, after)
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified subnet. You must terminate all running instances in the subnet before you can delete the
    subnet.

    Args:
        name(Text): The ID of the subnet or a name, ID, or JMES search path to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.subnet.absent:
                - name: value
                - subnet_id: value
    """

    result = dict(comment="", changes=None, name=name, result=True)
    ret = await hub.exec.boto3.client.ec2.subnet.search(ctx, jmes_path=name)
    if ret["result"]:
        # name is now the first id that matched the JMES search path
        name = ret["ret"]

    resource = hub.tool.boto3.resource.create(ctx, "ec2", "Subnet", name)

    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' already absent"
    else:
        try:
            ret = await hub.exec.boto3.client.ec2.delete_subnet(
                ctx, DryRun=ctx.test, **{"SubnetId": name}
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result["changes"] = differ.deep_diff(before, after)
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_subnets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe Subnets {ret['comment']}")
        return {}

    for subnet in ret["ret"]["Subnets"]:
        new_subnet = [
            {"vpc": subnet["VpcId"]},
            {"cidr_block": subnet["CidrBlock"]},
            {"tag_specifications": subnet.get("Tags", [])},
            {"availability_zone": subnet["AvailabilityZone"]},
            {"availability_zone_id": subnet["AvailabilityZoneId"]},
            {"outpost_arn": subnet.get("OutpostArn")},
        ]

        result[subnet["SubnetId"]] = {"aws.ec2.subnet.present": new_subnet}
        for i, data in enumerate(subnet["Ipv6CidrBlockAssociationSet"]):
            sub_subnet = copy.deepcopy(new_subnet)
            sub_subnet.append({"ipv6_cidrBlock": data["Ipv6CidrBlock"]})
            # The id needs to be in the name
            sub_subnet.append({"name": subnet["InstanceId"]})
            # Create a new state for this association set
            result[f"{subnet['SubnetId']}-{i}"] = {"aws.ec2.subnet.present": sub_subnet}
    return result



import shutil

# aws sts assume-role --role-arn arn:aws:iam::9999999:role --role-session-name "session name"
"""
#!/bin/bash
# Name: ~/.aws/assume.sh
# Usage: . ~/.aws/assume.sh arn:aws:iam::883373499178:role/xacct/developer
echo
echo " HISTCONTROL=ignoreboth"
echo " export AWS_DEFAULT_REGION=us-west-2"
round=1
for i in $(aws sts assume-role --role-session-name "${GOOGLE_USERNAME}" --role-arn "${1}"|jq -r '.Credentials|.AccessKeyId,.SecretAccessKey,.SessionToken,.Expiration')
do
    case $round in
        1)
        echo " export AWS_ACCESS_KEY_ID=$i"
        export AWS_ACCESS_KEY_ID="$i"
        round=$(( $round+1))
        ;;
        2)
        echo " export AWS_SECRET_ACCESS_KEY=$i"
        export AWS_SECRET_ACCESS_KEY="$i"
        round=$(( $round+1))
        ;;
        3)
        echo " export AWS_SESSION_TOKEN=$i"
        export AWS_SESSION_TOKEN="$i"
        round=$(( $round+1))
        ;;
        4)
        echo " export AWS_SESSION_EXPIRATION=$i"
        export AWS_SESSION_EXPIRATION="$i"
        ;;
    esac
done
unset AWS_PROFILE
echo
env|grep ^AWS
echo
aws sts get-caller-identity
OLDPS1="$PS1"
PS1="(`aws iam list-account-aliases|jq -r '.AccountAliases[0]'`) $PS1"
"""


def __virtual__(hub):
    path = shutil.which("aws")
    return bool(path), path


async def gather(hub):
    """
    Get profile information from node `aws sts`

    Example:
    .. code-block:: yaml

        aws.sts:
          default:
            # aws sts [operation]
            operation: assume-role
            # All other kwargs will be have "--" prepended and will be added to the operation call
            # See `aws sts [operation] help` for available kwargs
            role-arn: arn:aws:iam::99999999:role
            role-session-name: session_name
    """
    sub_profiles = {}
    for profile, ctx in hub.acct.PROFILES.get("aws.sts", {}).items():
        cmd = ["aws", "sts", ctx.pop("operation")] + [
            f"--{key} {value}" for key, value in ctx.items()
        ]
        # TODO run command and figure load all the profiles it defined
        hub.log.debug("aws sts authentication is not yet implemented")

    return sub_profiles



import shutil

# gsts --aws-role-arn arn:aws:iam::99999999999:role/xacct/developer --aws-profile=default --sp-id 99999999 --idp-id 999999999 --username foo@example.com


def __virtual__(hub):
    # npm install --global gsts
    path = shutil.which("gsts")
    return bool(path), path


async def gather(hub):
    """
    Get profile information from node gsts

    Example:
    .. code-block:: yaml

        aws.gsts:
          default:
            kwarg: thing
    """
    sub_profiles = {}
    for profile, ctx in hub.acct.PROFILES.get("aws.gsts", {}).items():
        hub.log.debug("GSTS authentication is not yet implemented")

    return sub_profiles



# Taken from salt 3000.4 salt/modules/boto_ssm.py


def unlock(hub):
    ...


'''
# -*- coding: utf-8 -*-
"""
Connection module for Amazon SSM

:configuration: This module uses IAM roles assigned to the instance through
    Instance Profiles. Dynamic credentials are then automatically obtained
    from AWS API and no further configuration is necessary. More Information
    available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

:depends: boto3
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging

import salt.utils.json as json

# Import Salt libs
import salt.utils.versions

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto libraries exist.
    """
    has_boto_reqs = salt.utils.versions.check_boto_reqs()
    if has_boto_reqs is True:
        __utils__["boto3.assign_funcs"](__name__, "ssm")
    return has_boto_reqs


def get_parameter(
        name,
        withdecryption=False,
        resp_json=False,
        region=None,
        key=None,
        keyid=None,
        profile=None,
):
    """
    Retrieves a parameter from SSM Parameter Store

    .. versionadded:: 3000

    .. code-block:: text

        salt-call boto_ssm.get_parameter test-param withdescription=True
    """
    conn = __utils__["boto3.get_connection"](
        "ssm", region=region, key=key, keyid=keyid, profile=profile
    )
    try:
        resp = conn.get_parameter(Name=name, WithDecryption=withdecryption)
    except conn.exceptions.ParameterNotFound:
        log.warning("get_parameter: Unable to locate name: %s", name)
        return False

    if resp_json:
        return json.loads(resp["Parameter"]["Value"])
    else:
        return resp["Parameter"]["Value"]


def put_parameter(
        Name,
        Value,
        Description=None,
        Type="String",
        KeyId=None,
        Overwrite=False,
        AllowedPattern=None,
        region=None,
        key=None,
        keyid=None,
        profile=None,
):
    """
    Sets a parameter in the SSM parameter store

    .. versionadded:: 3000

    .. code-block:: text

        salt-call boto_ssm.put_parameter test-param test_value Type=SecureString KeyId=alias/aws/ssm Description='test encrypted key'
    """
    conn = __utils__["boto3.get_connection"](
        "ssm", region=region, key=key, keyid=keyid, profile=profile
    )
    if Type not in ("String", "StringList", "SecureString"):
        raise AssertionError("Type needs to be String|StringList|SecureString")
    if Type == "SecureString" and not KeyId:
        raise AssertionError("Require KeyId with SecureString")

    boto_args = {}
    if Description:
        boto_args["Description"] = Description
    if KeyId:
        boto_args["KeyId"] = KeyId
    if AllowedPattern:
        boto_args["AllowedPattern"] = AllowedPattern

    try:
        resp = conn.put_parameter(
            Name=Name, Value=Value, Type=Type, Overwrite=Overwrite, **boto_args
        )
    except conn.exceptions.ParameterAlreadyExists:
        log.warning(
            "The parameter already exists."
            " To overwrite this value, set the Overwrite option in the request to True"
        )
        return False
    return resp["Version"]


def delete_parameter(Name, region=None, key=None, keyid=None, profile=None):
    """
    Removes a parameter from the SSM parameter store

    .. versionadded:: 3000

    .. code-block:: text

        salt-call boto_ssm.delete_parameter test-param
    """
    conn = __utils__["boto3.get_connection"](
        "ssm", region=region, key=key, keyid=keyid, profile=profile
    )
    try:
        resp = conn.delete_parameter(Name=Name)
    except conn.exceptions.ParameterNotFound:
        log.warning("delete_parameter: Unable to locate name: %s", Name)
        return False
    if resp["ResponseMetadata"]["HTTPStatusCode"] == 200:
        return True
    else:
        return False
'''



import re
from typing import Any
from typing import Dict
from typing import Generator

import boto3.session

ITERATION_FINISHED = object()

__func_alias__ = {"exec_": "exec"}


async def exec_(
    hub,
    ctx,
    service_name: str,
    operation: str,
    jmes_search_path: str = None,
    *op_args,
    **op_kwargs: Dict[str, Any],
) -> Any:
    """
    :param hub:
    :param ctx:
    :param service_name: The name of the service client to create
    :param operation: The operation to run from the service client
    :param jmes_search_path: The JMES path to use as a filter for paginated results
    :param op_args: arguments to pass to the operation call
    :param op_kwargs: keyword arguments to pass to the operation call

    :return: The result of the operation call
    """
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(service_name=service_name, **ctx.acct)

    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in op_kwargs.items() if v is not None}

    can_paginate = client.can_paginate(operation)

    if can_paginate:
        hub.log.debug(f"Paginating results for {service_name}.{operation}")
        paginator = client.get_paginator(operation)
        pages = paginator.paginate(*op_args, **kwargs)
        if jmes_search_path is None:
            return await hub.pop.loop.wrap(pages.build_full_result)
        else:
            iterator = await hub.pop.loop.wrap(pages.search, jmes_search_path)
            return [_ for _ in iterator]
    else:
        hub.log.debug(f"Getting raw results for {service_name}.{operation}")
        op = getattr(client, operation)
        return await hub.pop.loop.wrap(op, *op_args, **kwargs)


async def wait(
    hub, ctx, service_name: str, waiter_name: str, *wt_args, **wt_kwargs
) -> None:
    """
    Asynchronously wait for the named resource to be avilable

    :param hub:
    :param ctx:
    :param service_name: The name of the service client to retrieve
    :param waiter_name: The name of the waiter to get from the service client
    :param wt_args: Args to pass to the wait function
    :param wt_kwargs: kwargs to pass to the wait function
    """
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(service_name=service_name, **ctx.acct)

    if waiter_name not in client.waiter_names:
        raise NameError(
            f"No waiter '{waiter_name}'. "
            f"Available waiters for '{service_name}' are: {' '.join(client.waiter_names)}"
        )

    waiter = client.get_waiter(waiter_name)

    await hub.pop.loop.wrap(waiter.wait, *wt_args, **wt_kwargs)


async def search(
    hub, ctx, service_name: str, collection: str, jmes_path: str = "*[]"
) -> str:
    """
    :param hub:
    :param ctx:
    :param service_name:
    :param collection:
    :param jmes_path:
    :return:
    """
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(service_name=service_name, **ctx.acct)

    # Find a paginator that can describe this collection
    for operation in client.meta.method_to_api_mapping:
        if client.can_paginate(operation) and re.match(
            f"[a-z]+_{collection}s?$", operation
        ):
            break
    else:
        raise AttributeError(
            f"Could not find a paginator for {service_name}.{collection}"
        )

    paginator = client.get_paginator(operation)
    pages = paginator.paginate()
    iterator: Generator = await hub.pop.loop.wrap(pages.search, jmes_path)
    ret = next(iterator, None)
    iterator.close()
    return ret



from typing import Any
from typing import Dict


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Get profile names from encrypted AWS credential files.
    Standardize the profile keys to what a boto3 Client requires

    Example:
    .. code-block:: yaml

        aws.boto:
          profile_name:
            id: XXXXXXXXXXXXXXXXX
            key: XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            region: us-east-1
    """
    sub_profiles = {}
    for profile, ctx in profiles.get("aws.boto", {}).items():
        # Add a boto session to the ctx for exec and state modules
        # Strip any args that were used for authentication
        # Boto uses the default account if none was specified
        sub_profiles[profile] = ctx
    return sub_profiles



from typing import Any
from typing import Dict

import botocore.config


def sig_gather(hub, profiles) -> Dict[str, Any]:
    ...


def post_gather(hub, ctx) -> Dict[str, Any]:
    """
    TODO Sanitize these profiles and make sure they have everything needed
    https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """
    profiles = ctx.ret or {}

    hub.log.info(f"Read {len(profiles)} profiles from {ctx.ref}")

    for name, profile in profiles.items():
        profiles[name] = _sanitize_profile(profile)

    return profiles


def _sanitize_profile(profile: Dict[str, str]) -> Dict[str, str]:
    return dict(
        region_name=_key_options(profile, "region_name", "region", "location"),
        api_version=profile.get("api_version"),
        verify=profile.get("verify"),
        use_ssl=profile.get("use_ssl"),
        endpoint_url=_key_options(profile, "endpoint_url", "endpoint"),
        aws_access_key_id=_key_options(profile, "aws_access_key_id", "key_id", "id"),
        aws_secret_access_key=_key_options(
            profile, "aws_secret_access_key", "secret_access_key", "access_key", "key"
        ),
        aws_session_token=_key_options(
            profile, "aws_session_token", "session_token", "token", "key"
        ),
        config=botocore.config.Config(**profile.get("config", {})),
    )


def _key_options(d: Dict[str, Any], *keys):
    for key in keys:
        if key in d:
            return d[key]
    else:
        return None
